//
// C++ Implementation: porcelainbutton
//
// Description:
//
//
// Author: Remi Villatel <maxilys@tele2.fr>, (C) 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <qtooltip.h>
#include <qpainter.h>
#include <qpixmap.h>

#include "enums.h"
#include "porcelainbutton.h"
#include "porcelainhandler.h"
#include "pixmaps.h"

#include "embeddata.h"

PorcelainButton::PorcelainButton(PorcelainClient *parent, const char *name, const QString& tip, ButtonType type) : QButton(parent->widget(), name), m_client(parent), m_lastMouse(0), m_type(type), hover(false), isOnAllDesktops(false), isMaximized(false)
{
	QToolTip::add(this, tip);
	setCursor(arrowCursor);
	setBackgroundMode(NoBackground);
	int titleSize = PorcelainHandler::titleSize();
	// int borderSize = (titleSize/4)-1;
	int buttonSize = (titleSize/2)+6;
	setFixedSize(buttonSize, buttonSize);
}


PorcelainButton::~PorcelainButton()
{
}

void PorcelainButton::setTipText(const QString &tip)
{
	QToolTip::remove(this);
	QToolTip::add(this, tip);
}

QSize PorcelainButton::sizeHint() const
{
	int titleSize = PorcelainHandler::titleSize();
	// int borderSize = (titleSize/4)-1;
	int buttonSize = (titleSize/2)+6;
	return QSize(buttonSize, buttonSize);
}

void PorcelainButton::enterEvent(QEvent *e)
{
	hover = true;
	repaint(false);
	QButton::enterEvent(e);
}

void PorcelainButton::leaveEvent(QEvent *e)
{
	hover = false;
	repaint(false);
	QButton::enterEvent(e);
}

void PorcelainButton::mousePressEvent(QMouseEvent *e)
{
	m_lastMouse = e->button();
	QMouseEvent me(e->type(), e->pos(), e->globalPos(), LeftButton, e->state());
	QButton::mousePressEvent(&me);
}

void PorcelainButton::mouseReleaseEvent(QMouseEvent *e)
{
	m_lastMouse = e->button();
	QMouseEvent me(e->type(), e->pos(), e->globalPos(), LeftButton, e->state());
	QButton::mouseReleaseEvent(&me);
}

void PorcelainButton::drawButton(QPainter *painter)
{
	if (!PorcelainHandler::initialized())
		return;

	bool active = m_client->isActive();
	int titleSize = PorcelainHandler::titleSize();
	// int borderSize = (titleSize/4)-1;
	int buttonSize = (titleSize/2)+6;

	int ofx = (buttonSize-16)/2;
	int ofy = ofx-1;

	// MXLS
	QColorGroup widgetGroup;
	widgetGroup = m_client->widget()->colorGroup();
	QColor coground = widgetGroup.background();

	QPixmap pufferPixmap;
	pufferPixmap.resize(buttonSize, buttonSize);

	QPainter pufferPainter(&pufferPixmap);
	pufferPixmap.fill(coground);

	if (m_type == ButtonMenu)
	{
		QPixmap menu_icon = m_client->icon().pixmap(QIconSet::Small, QIconSet::Normal);
		pufferPainter.drawPixmap(ofx, ofy, menu_icon);
	}
	else
	{
		// Paint the little icon on the buttons.
		// Lotsa trouble with this! Stupid scope!

		if (hover && !isDown())
		{
			// Hover
			if (m_type == ButtonHelp)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_help_hover : *Pixmaps::inactive_help_hover);
			else if (m_type == ButtonMax)
			{
				if (isMaximized)
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_restore_hover : *Pixmaps::inactive_restore_hover);
				else
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_max_hover : *Pixmaps::inactive_max_hover);
			}
			else if (m_type == ButtonMin)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_min_hover : *Pixmaps::inactive_min_hover);
			else if (m_type == ButtonClose)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_close_hover : *Pixmaps::inactive_close_hover);
			else if (m_type == ButtonOnAllDesktops)
			{
				if (isOnAllDesktops)
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_unsticky_hover : *Pixmaps::inactive_unsticky_hover);
				else
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_sticky_hover : *Pixmaps::inactive_sticky_hover);
			}
			else if (m_type == ButtonAbove)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_above_hover : *Pixmaps::inactive_above_hover);
			else if (m_type == ButtonBelow)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_below_hover : *Pixmaps::inactive_below_hover);
		}
		else if (isDown())
		{
			// Sunken
			if (m_type == ButtonHelp)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_help_sunken : *Pixmaps::inactive_help_sunken);
			else if (m_type == ButtonMax)
			{
				if (isMaximized)
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_restore_sunken : *Pixmaps::inactive_restore_sunken);
				else
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_max_sunken : *Pixmaps::inactive_max_sunken);
			}
			else if (m_type == ButtonMin)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_min_sunken : *Pixmaps::inactive_min_sunken);
			else if (m_type == ButtonClose)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_close_sunken : *Pixmaps::inactive_close_sunken);
			else if (m_type == ButtonOnAllDesktops)
			{
				if (isOnAllDesktops)
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_unsticky_sunken : *Pixmaps::inactive_unsticky_sunken);
				else
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_sticky_sunken : *Pixmaps::inactive_sticky_sunken);
			}
			else if (m_type == ButtonAbove)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_above_sunken : *Pixmaps::inactive_above_sunken);
			else if (m_type == ButtonBelow)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_below_sunken : *Pixmaps::inactive_below_sunken);
		}
		else
		{
			// Normal
			if (m_type == ButtonHelp)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_help_normal : *Pixmaps::inactive_help_normal);
			else if (m_type == ButtonMax)
			{
				if (isMaximized)
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_restore_normal : *Pixmaps::inactive_restore_normal);
				else
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_max_normal : *Pixmaps::inactive_max_normal);
			}
			else if (m_type == ButtonMin)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_min_normal : *Pixmaps::inactive_min_normal);
			else if (m_type == ButtonClose)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_close_normal : *Pixmaps::inactive_close_normal);
			else if (m_type == ButtonOnAllDesktops)
			{
				if (isOnAllDesktops)
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_unsticky_normal : *Pixmaps::inactive_unsticky_normal);
				else
					pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_sticky_normal : *Pixmaps::inactive_sticky_normal);
			}
			else if (m_type == ButtonAbove)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_above_normal : *Pixmaps::inactive_above_normal);
			else if (m_type == ButtonBelow)
				pufferPainter.drawImage(ofx, ofy, active ? *Pixmaps::active_below_normal : *Pixmaps::inactive_below_normal);
		}
	}
	pufferPainter.end();
	painter->drawPixmap(0, 0, pufferPixmap);
}
#include "porcelainbutton.moc"
